/*******************************************************************************
  *
  *  program     : AbstractGraphObject.java
  *  author      : Helge Janicke
  *  version     : 2001-01-20(001)
  *  (c)         : Helge Janicke, Karsten Wolke, Niels-Peter de Witt
  *  email       : heljanic@agent.fho-emden.de
  *  description : AbstractGraphObject definiert Eigenschaften, die notwendig
  *									sind um ein Objekt in einem LayerPainter darstellen zu
  * 							  knnen.
  ******************************************************************************
  * modifications:
  * date: ******* who: ****** description: *************************************
  * ----------------------------------------------------------------------------
  * 2000-10-19    heljanic
  * ----------------------------------------------------------------------------
  ******************************************************************************
  * comments:
  *
  ******************************************************************************
  * license:
  *
*******************************************************************************/
package jdraw;

import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Color;
import java.awt.Paint;
import java.awt.Stroke;
import java.awt.BasicStroke;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Point2D;
import java.awt.geom.PathIterator;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.geom.CubicCurve2D;
import java.awt.geom.QuadCurve2D;
import java.awt.Color;
import java.awt.BasicStroke;


/** Definiert alle Methoden, die notwendig sind um ein Objekt in einem LayerPainter
  * darstellen zu knnen.
  */
public abstract class AbstractGraphObject implements Shape, Selectable {

  final static Paint DEFAULT_LINEPAINT = Color.black;
  final static Paint DEFAULT_FILLPAINT = Color.blue;
  final static Stroke DEFAULT_STROKE = new BasicStroke();

  // Methoden, die von Shape und Selectable definiert werden.
  public abstract boolean contains(double x, double y);
  public abstract boolean contains(double x, double y, double w, double h);
  public abstract boolean contains(Point2D p);
  public abstract Rectangle getBounds();
  public abstract Rectangle2D getBounds2D();
  public abstract boolean intersects(double x, double y, double w, double h);
  public abstract boolean intersects(Rectangle2D r);

  // Methoden die von Shape definiert werden.
  public abstract PathIterator getPathIterator(AffineTransform at);
  public abstract PathIterator getPathIterator(AffineTransform at, double flatness);

  public abstract AffineTransform getAffineTransform();
  public abstract void setAffineTransform(AffineTransform at);

  /** Liefert das Paint Objekt, das zum Fllen des AbstractGraphObjects verwendet wird.
    * @return Das Paint Objekt zum Fllen.
    */
  public abstract Paint getFillPaint();

	/** Setzt das Paint Objekt, das zum Fllen des AbstractGraphObjects verwendet wird.
	  * @param p Das neue Paint Object.
	  */
  public abstract void setFillPaint(Paint p);

  /** Liefert das Paint Objekt, das fr die Umrandung des AbstractGraphObjects verwendet
    * wird.
    * @return Das Paint Objekt fr die Umrandung.
    */
  public abstract Paint getLinePaint();

	/** Setzt das Paint Objekt, das fr die Umrandung des AbstractGraphObjects verwendet wird.
	  * @param p Das neue Paint Object.
	  */
  public abstract void setLinePaint(Paint p);

  /** Liefert die Strichart des AbstractGraphObjects.
    * @return Das zum Zeichnen der Umrandung verwendete Stroke Objekt.
    */
  public abstract Stroke getStroke();

	/** Setzt die Strichart, die zum Zeichnen des AbstractGraphObjects verwendet wird.
	  * @param s Das Stroke Objekt, das verwendet werden soll.
	  */
  public abstract void setStroke(Stroke s);

  /** True wenn das AbstractGraphObject gefllt werden soll.
    * @return <tt>true</tt> wenn gefllt, sonst <tt>false</tt>.
    */
  public abstract boolean isFilled();

	/** Schaltet die Fllung des Objektes ein und aus.
		* @param flag <tt>true</tt> bedeutet gefllt.
		*/
  public abstract void setFilled(boolean flag);

  /** True wenn die Umrandung des AbstractGraphObject gezeichnet werden soll.
    * @return <tt>true</tt> wenn umrandet, sonst <tt>false</tt>.
    */
  public abstract boolean isOutlined();

	/** Schaltet die Umrandung des Objektes ein und aus.
	  * @param flag <tt>true</tt> bedeutet umrandet.
	  */
  public abstract void setOutlined(boolean flag);

  /** Prft ob das Rechteck <tt>r</tt> die Umrandung des Shapes schneidet.
    * Dazu wird der PathIterator des Shapes durchlaufen und jede Kurve auf
    * einen Schnitt geprft.
    * @return <tt>true</tt> wenn mindestens eine Kurve geschnitten wurde, sonst
    *         <tt>false</tt>.
    */
  public static boolean intersectsBoundary(Shape s, Rectangle2D r) {
    double lastX=0;
    double lastY=0;
    double[] points = new double[6];
    Shape testshape = null;
    PathIterator pi = s.getPathIterator(null);

    if (!pi.isDone()) {

      // get starting point to last? doubles.
      // Request: pathiterator starts with a SEG_MOVE or SEG_CLOSE
      switch (pi.currentSegment(points)) {
        case PathIterator.SEG_CLOSE: break;
        case PathIterator.SEG_MOVETO:
          // returns one point.
          lastX = points[0];
          lastY = points[1];
          break;
        default:
          System.err.println("path iterator didn't start with move!");
          break;
      }

      // processing the iterator.
      // Generating proper shapes for each segment.
      while (!pi.isDone()) {
        switch (pi.currentSegment(points)) {
          case PathIterator.SEG_CLOSE: break;
          case PathIterator.SEG_MOVETO:
            // returns one point.
            lastX = points[0];
            lastY = points[1];
            break;
          case PathIterator.SEG_LINETO:
            // returns one point.
            testshape = new Line2D.Double(lastX, lastY, points[0], points[1]);
            lastX = points[0];
            lastY = points[1];
            break;
          case PathIterator.SEG_QUADTO:
            // returns two points.
            testshape = new QuadCurve2D.Double(lastX, lastY, points[0], points[1], points[2], points[3]);
            lastX = points[2];
            lastY = points[3];
            break;
          case PathIterator.SEG_CUBICTO:
            // returns three points.
            testshape = new CubicCurve2D.Double(lastX, lastY, points[0], points[1], points[2], points[3], points[4], points[5]);
            lastX = points[4];
            lastY = points[5];
            break;
          default:
            System.err.println("path iterator: not processed returnValue");
            testshape=null;
            break;
        }

        // getting the next segment.
        pi.next();

        // check the applied testshape for intersection with rectangle r.
        if (testshape != null && testshape.intersects(r)) return true;
      }
    }
    return false;
  }

}
