/*******************************************************************************
  *
  *  program     : DefaultGlassPane.java
  *  author      : Helge Janicke
  *  version     : 2000-10-19(001)
  *  (c)         : Helge Janicke, Karsten Wolke, Niels-Peter de Witt
  *  email       : heljanic@agent.fho-emden.de
  *  description : Ist eine einfache Implementierung einer GlassPane.
  *
  ******************************************************************************
  * modifications:
  * date: ******* who: ****** description: *************************************
  * ----------------------------------------------------------------------------
  * 2000-10-19    heljanic    your description here
  * ----------------------------------------------------------------------------
  ******************************************************************************
  * comments:
  *
  ******************************************************************************
  * license:
  *
*******************************************************************************/

package jdraw;

import java.awt.Paint;
import java.awt.Stroke;
import java.awt.Shape;

import java.util.Vector;
import javax.swing.*;
import java.awt.*;
import java.awt.Graphics2D;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Component;
import java.awt.geom.AffineTransform;
import java.awt.AlphaComposite;

/** Eine GlassPane dient dazu die ausgewhlten Zeichenobjekte darzustellen,
  * sowie die im Zeichenvorgang befindlichen Objekte anzuzeigen.
  */
public class DefaultGlassPane extends JPanel implements GlassPane {
  /** Hier werden die selektierten Zeichenobjekte gespeichert. */
  private Vector selection = new Vector();

  /** Hier werden die im Zeichnen befindlichen Objekte gespeichert.*/
  private Vector shapes = new Vector();

  /** Dieses Objekt enthlt ein Selectable Objekt, das die aktuelle Auswahl
    * reprsentiert.
    */
  private Selectable selectable = null;

  /** Hier wird das Component Objekt gespeichert, dessen repaint Routine bei
    * einem refresh() aufgerufen wird.*/
  private Component refreshComponent = null;

  /** Das Panel, das die Informationen des Infoable Objektes enthlt. */
  private JPanel panInfo;

  /** Dieses Panelmerkt sich das letzte benutzte Infopanel. */
  private JPanel lastInfo = new JPanel();

  /** Erzeugt eine DefaultGlassPane */
  public DefaultGlassPane() {
  }

  /** Zeichnet die refreshComponent neu und aktualisiert das InfoPanel. */
  public void refresh() {
    JPanel newInfo;
    if (refreshComponent != null) refreshComponent.repaint();
    if (getSelectable() instanceof Infoable) {
      Infoable inf = (Infoable) getSelectable();
      if (panInfo != null) {
        newInfo = inf.getInfoPanel();
        if (!lastInfo.equals(newInfo)) {
          panInfo.removeAll();
          panInfo.add(newInfo, BorderLayout.WEST);
          lastInfo = newInfo;
        }
      }
      if (getSelectable() instanceof AbstractGraphObject)
        inf.updateInfoPanel((AbstractGraphObject) getSelectable());
    } else {
      if (panInfo != null) panInfo.removeAll();
      lastInfo = new JPanel();
    }
  }

  /** Setzt das Component Objekt, dessen repaint() Methode bei einem
    * refresh() aufgerufen wird.
    * @param c das Component Objekt.
    */
  public void setRefreshComponent(Component c) {
    this.refreshComponent = c;
  }

  /** Fgt ein Selectable Objekt zur aktuellen Auswahl hinzu.
    * Dieses Selectable wird in den selection Vector eingetragen.
    * Falls es mehr als eine Auswahl geben sollte wird ein neues
    * selectable (Eine Instanz von ImplicitGroup.java) generiert.
    * @param o Das Selectable, da der Auswahl hinzugefgt werden soll.
    */
  public void addSelection(Selectable o) {
    selection.add(o);
    this.selectable = createImplicitGroup();
  }

  /** Fgt der Auswahl eine neue Farbe hinzu. Alle nun folgenden Selectable
    * Objekte werden in dieser Farbe dargestellt.
    * @param p Das Paint Objekt mit dem folgende Selectables dargestellt werden sollen.
    */
  public void addSelection(Paint p) {
    selection.add(p);
  }

  /** Fgt der Auswahl einen neuen Linienstiel hinzu. Alle nun folgenden Selectable
    * Objekte werden mit diesem Linienstil dargestellt.
    * @param s Das Stroke Objekt mit dem folgende Selectables dargestellt werden sollen.
    */
  public void addSelection(Stroke s) {
    selection.add(s);
  }

  /** Entfernt ein Selectable aus der aktuellen Auswahl.
    * Falls dieses Objekt ein Teil einer mehrfachen Selektion war, wird es aus dem entsprechenden ImplicitGroup Objekt entfernt.
    * @param o Das Selectable, das entfernt werden soll.
    */
  public void remSelection(Selectable o) {
    if (getSelectable() instanceof Transformable) {
      ((Transformable) getSelectable()).freeAccessPoints();
    }
    selection.remove(o);
    this.selectable = createImplicitGroup();
  }

  /** Entfernt ein bestehendes Paint Objekt aus der selection Liste.
    * @param p Das Paint Objekt, das entfernt werden soll.
    */
  public void remSelection(Paint p) {
    selection.remove(p);
  }

  /** Entfernt ein bestehendes Stroke Objekt aus der selection Liste.
    * @param s Das Stroke Objekt, das entfernt werden soll.
    */
  public void remSelection(Stroke s) {
    selection.remove(s);
  }

  /** Entfernt die gesamte Auswahl. */
  public void clearSelection() {
    if (getSelectable() instanceof Transformable) {
      ((Transformable) getSelectable()).freeAccessPoints();
    }
    selection.clear();
    this.selectable = null;
  }

  /** Prft ob ein Selectable tatschlich ausgewhlt ist.
    * @param o Das Selectable, das geprft werden soll.
    * @return <tt>true</tt> wenn das Selectable tatschlich ausgewhlt ist.
    */
  public boolean isSelected(Selectable o) {
    return (selection.indexOf(o) >= 0);
  }

  /** Fgt ein Shape zur Liste der im Zeichnen befindlichen Objekte hinzu.
    * @param s Das Shape, das gezeichnet werden soll.
    */
  public void addShape(Shape s) {
    shapes.add(s);
  }

  /** Entfernt ein Shape aus der Liste der im Zeichnen befindlichen Objekte.
    * @param s Das Shape, das entfernt werden soll.
    */
  public void remShape(Shape s) {
    shapes.remove(s);
  }

  /** Entfernt alle Shapes aus der Liste der im Zeichnen befindlichen Objekte.*/
  public void clearShapes() {
    shapes.clear();
  }

  /** Liefert eine Liste aller im selection Vektor befindlichen Objekte zurck.
    * Diese Liste enhlt Selectable, Strokes und Paint Objekte.
    * @return Einen Array, der alle ausgewhlten Ojekte enthlt.
    */
  public Object[] getSelection() {
    return selection.toArray();
  }

  /** Erzeugt aus allen Selectables, die es findet ein ImplicitGroup Objekt.
   * Dieses wird nun als Selectable betrachtet und stellt evtl. die AccessPoints bereit...
   * Falls nur ein Selectable gruppiert werden soll, so wird kein ImplicitGroup Objekt
   * zurckgegeben, sondern das Selectable selbst.
   * @return Ein Selectable, das die aktuelle Auswahl reprsentiert.
   */
  private Selectable createImplicitGroup() {
    ImplicitGroup grp = new ImplicitGroup();
    Selectable first = null;
    boolean createGroup=false;
    for (int i=0; i<selection.size(); i++) {
      if (selection.get(i) instanceof Selectable) {
        if (first==null) {
          first = (Selectable) selection.get(i);
          grp.add(first);
        } else {
          createGroup = true;
          grp.add((Selectable) selection.get(i));
        }
      }
    }

    if (createGroup) return grp;
    return first;
  }

  /** Liefert die aktuelle Auswahl.
    * Falls es sich um eine mehrfache Auswahl handelt, so wird ein
    * ImplicitGroup Objekt zurckgegeben, da die einzelnen Objekte kapselt.
    * @return Die Aktuelle Auswahl.
    */
  public Selectable getSelectable() {
    return this.selectable;
  }

  /** Zeichnet alle Shape objecte, die sich in dem Vector shapes befinden.
   * befindet sich im Vector ein Paint oder ein Stroke oder ein AffineTransform
   * -Object, so wird es gesetzt.
   */
  public void paint (Graphics g) {
    Graphics2D g2  = (Graphics2D) g;
    Shape selShape = null;
    Shape drgShape = null;
    Shape defShape = null;
    AbstractAccessPoint[] ap = null;
    Object obj = null;

    g2.setComposite(AlphaComposite.Src);


    for (int i=0; i<selection.size(); i++) {
      obj = selection.get(i);

      if (obj instanceof Paint) {
        // das Paint setzen...
        g2.setPaint((Paint) obj);
      } else if (obj instanceof Stroke) {
        // den Stroke setzen...
        g2.setStroke((Stroke) obj);
      } else if (obj instanceof Selectable) {
        // Eine Selektion... Ein default Shape generieren, wenn mglich...
        defShape = (obj instanceof Shape)? (Shape) obj : null;

        // Ohh, es kann sogar transformiert werden....
        if(obj instanceof Transformable) {
          // Das selected bzw. dragged Shape holen.
          Transformable trans = ((Transformable) obj);
          selShape = trans.supportSelectedShape();
          drgShape = trans.supportDragShape();
          // Die AccessPoints holen.
          // ap = trans.getAccessPoints();
        }

        // affine Transformation ausfhren.
        AffineTransform old = g2.getTransform();

        if (selShape != null) {
          // es gibt ein selectedShape, also zeichnen wir dieses
          g2.draw(selShape);
        } else if (defShape != null) {
          // es gibt kein selectedShape, also das defaultShape zeichnen.
          g2.draw(defShape);
        } else {
          // Ohh je auch kein defaultShape... nun ja dann eben die Boundingbox!
          g2.draw(((Selectable) obj).getBounds2D());
        }
        if (drgShape != null) {
          // Es wird sogar ein draggedShape untersttzt, whow!
          g2.draw(drgShape);
        }

        g2.setTransform(old);

        Selectable selec = getSelectable();

        if (selec instanceof Transformable) {
          Transformable trans = (Transformable) selec;
          ap = trans.getAccessPoints();
          Shape drgShp = trans.supportDragShape();
          Shape selShp = trans.supportSelectedShape();
          if (drgShp != null) g2.draw(drgShp);
          if (selShp != null) {
            g2.setPaint(Color.yellow);
            g2.draw(selShp);
          }
        } else if (selec instanceof Shape) {
          g2.draw((Shape) selec);
        }

        if (ap != null) {
          // die AccessPoints zeichnen...
          for (int j=0; j<ap.length; j++) {
            g2.setPaint(ap[j].getPaint());
            g2.setStroke(ap[j].getStroke());
            if (ap[j].isFilled()) {
              g2.fill(ap[j].getAppearance());
            } else {
              g2.draw(ap[j].getAppearance());
            }
          }
        }
      }
    }
    // System.err.println("try drawing shapes");
    for (int j=0; j<shapes.size(); j++) {
      try {
        g2.setPaint(Color.red);
        g2.draw((Shape) shapes.get(j));
      } catch (ClassCastException exc) {
        System.err.println("There was something different in the shapes Vector than a Shape!!!");
      }
    }
  }

  /** Setzt das aktuelle Infopanel, das angezeigt werden soll.
    * @param infopanel Das Infopanel, das angezeigt werden soll.
    */
  public void setInfoPanel(JPanel infopanel) {
    this.panInfo = infopanel;
    this.panInfo.setLayout(new BorderLayout() );
    // System.err.println("hello setInfoPanel");
    // panInfo.add(new javax.swing.JLabel("testlabel"));
  }

}