/*******************************************************************************
  *
  *  program     : GraphGroup.java
  *  author      : Helge Janicke
  *  version     : 2000-10-19(001)
  *  (c)         : Helge Janicke, Karsten Wolke, Niels-Peter de Witt
  *  email       : heljanic@agent.fho-emden.de
  *  description :
  ******************************************************************************
  * modifications:
  * date: ******* who: ****** description: *************************************
  * ----------------------------------------------------------------------------
  * 2000-10-19    heljanic
  * ----------------------------------------------------------------------------
  ******************************************************************************
  * comments:
  *
  ******************************************************************************
  * license:
  *
*******************************************************************************/
package jdraw;

import java.awt.*;
import java.awt.geom.*;
import javax.swing.tree.*;
import java.util.*;

/** Ermglicht eine Gruppierung von Grafischen Objekten und beliebig vielen Gruppen.
 * Ein Objekt dieser Klasse, das an einen Gruppierungsknoten im Baum eingehngt wird
 * ruft rekursiv alle Grafischen Objekte/ Gruppen der Unterknoten auf.
 */

public class GraphGroup implements Selectable {
  private Rectangle bounds = new Rectangle();
  private DefaultMutableTreeNode node = null;

  /** Initialisiert die neue Gruppe.
   * @param node der Knoten, an dem dieses Objekt in den Baum gehngt wird.
   */
  public GraphGroup(DefaultMutableTreeNode node) {
    this.node = node;
  }

/** Berechnet das umschlieende Rechteck.
 * @return das Rechteck.
 */
  private Rectangle computeBounds() {
    double min_x, min_y, max_x, max_y;

    // Liste aller Kindknoten einlesen und in ein Selectable
    Enumeration enum = node.children();
    if (enum.hasMoreElements()) {
      Selectable firstChild = (Selectable) ((DefaultMutableTreeNode) (enum.nextElement())).getUserObject();

      // Die Grenzen des ersten Kindes als minimale Grenzen setzen.
      if (firstChild != null) bounds = firstChild.getBounds();
      min_x = bounds.getMinX();
      min_y = bounds.getMinY();
      max_x = bounds.getMaxX();
      max_y = bounds.getMaxY();


      // alle
      while (enum.hasMoreElements()) {
        try {
          Selectable child = (Selectable) ((DefaultMutableTreeNode) (enum.nextElement())).getUserObject();
          if (child != null) {
            Rectangle rect = child.getBounds();
            if (rect.getMinX() < min_x ) min_x = rect.getMinX();
            if (rect.getMinY() < min_y ) min_y = rect.getMinY();
            if (rect.getMaxX() > max_x ) max_x = rect.getMaxX();
            if (rect.getMaxY() > max_y ) max_y = rect.getMaxY();
          }
        } catch (ClassCastException exc) {
          System.err.println("Object in Group wasn't a Selectable.");
          exc.printStackTrace();
        }
      }
      bounds.setBounds((int) min_x, (int) min_y, (int) (max_x - min_x), (int) (max_y - min_y));
    }
    return bounds;
  }

/** Fhrt die Transformation rekursiv auf alle Objekte dieser Gruppe aus.
 * @param at Die Affine Transformation, die angewendet werden soll.
 */
  public void transform(AffineTransform at) {
  // TODO: implement recursiv call of all subcomponents.
  }

/** Berechnet das umschlieende Rechteck.
 * @return das Rechteck.
 */  public Rectangle getBounds() {
    computeBounds();
    return bounds;
  }

/** Berechnet das umschlieende Rechteck.
 * @return das Rechteck.
 */  public Rectangle2D getBounds2D() {
    computeBounds();
    return bounds;
  }

/** Prft ob der Punkt innehalb der Grenzen liegt.
 * @param x die x-Koordinate.
 * @param y die y-Koordinate.
 */
  public boolean contains(double x, double y) {
    return getBounds().contains(x,y);
  }

/** Prft ob der Punkt innerhalb der Grenzen liegt.
 * @param p der Punkt.
 */
  public boolean contains(Point2D p) {
    return getBounds().contains(p);
  }

/** Prft das Rechteck mit mindestens einem Midglied der Gruppe schneidet.
 * @param x die x-Koordinate.
 * @param y die y-Koordinate.
 * @param w die Breite.
 * @param h die Hhe.
 */
  public boolean intersects(double x, double y, double w, double h) {
    boolean rv = false;
    for (Enumeration enum = node.children(); enum.hasMoreElements(); ) {
      try {
        Selectable sel = (Selectable) ((DefaultMutableTreeNode) (enum.nextElement())).getUserObject();
        if (sel != null) {
           // rv = rv | sel.intersects(x,y,w,h);
           // substitute with faster implementation 08.06.2000 by Helge Janicke
           if (sel.intersects(x,y,w,h)) return true;
        }
      } catch (ClassCastException exc) {
        System.err.println("Object in Group wasn't a Selectable.");
        exc.printStackTrace();
      }
    }
    return rv;
  }

/** Prft das Rechteck mit mindestens einem Mitglied der Gruppe schneidet.
 * @param r das Rechteck.
 */
  public boolean intersects(Rectangle2D r) {
    boolean rv = false;
    for (Enumeration enum = node.children(); enum.hasMoreElements(); ) {
      try {
        Selectable sel = (Selectable) ((DefaultMutableTreeNode) (enum.nextElement())).getUserObject();
        if (sel != null) {
           // rv = rv | sel.intersects(x,y,w,h);
           // substitute with faster implementation 08.06.2000 by Helge Janicke
           if (sel.intersects(r)) return true;
        }
      } catch (ClassCastException exc) {
        System.err.println("Object in Group wasn't a Selectable.");
        exc.printStackTrace();
      }
    }
    return rv;
  }

/** Prft ob ein Rechteck innhalb der Grenzen liegt.
 * @param x die x-Koordinate.
 * @param y die y-Koordinate.
 * @param w die Breite.
 * @param h die Hhe.
 */
  public boolean contains(double x, double y, double w, double h) {
    return getBounds().contains(x,y,w,h);
  }

/** Prft ob ein Rechteck innhalb der Grenzen liegt.
 * @param r das Rechteck.
 */
  public boolean contains(Rectangle2D r) {
    return getBounds().contains(r);
  }

/** Overwritten object.toString()
 * @return kurze Beschreibung.
 */
  public String toString() {return ("Group of " + node.getChildCount() + " items");}
}
