/*******************************************************************************
  *
  *  program     : ImplicitGroup.java
  *  author      : Helge Janicke
  *  version     : 2000-10-19(001)
  *  (c)         : Helge Janicke, Karsten Wolke, Niels-Peter de Witt
  *  email       : heljanic@agent.fho-emden.de
  *  description : Diese Klasse fasst eine mehrfache Auswahl zu einer losen
  *     Gruppierung zusammen.
  ******************************************************************************
  * modifications:
  * date: ******* who: ****** description: *************************************
  * ----------------------------------------------------------------------------
  * 2000-10-19    heljanic
  * ----------------------------------------------------------------------------
  ******************************************************************************
  * comments:
  *
  ******************************************************************************
  * license:
  *
*******************************************************************************/
package jdraw;

import java.awt.geom.AffineTransform;
import java.awt.Rectangle;
import java.awt.Paint;
import java.awt.geom.Point2D;
import java.awt.geom.PathIterator;
import java.awt.Stroke;
import java.awt.geom.Rectangle2D;
import java.awt.Color;
import java.util.*;
import java.awt.event.MouseEvent;
import java.awt.Shape;

/** Diese Klasse fasst eine mehrfache Auswahl zu einer losen Gruppierung zusammen
  * Alle Transformationen werden an die Gruppenmitglieder weitergereicht.
  */
public class ImplicitGroup extends SimpleGraphObject implements Transformable {

  /** Hier werden die Gruppenmitglierder gespeichert. */
  private Vector group = new Vector();

  /** Dies ist ein flag das zur Analyse der einzelnen Gruppenmitglieder bentigt wird. Es zeigt an ob das designierte Gruppenmitglied sich berhaupt gruppieren lt.*/
  private boolean support = true;

  /** Die BoundingBox der Gruppe. */
  private Rectangle2D bounds;

  /** Ein Punkt der fr Transformationen als Startpunkt gemerkt wird. */
  private Point2D startPoint;

  /** Die AccessPoints die fr Gruppentransformationen bereitgestellt werden.*/
  private SimpleAccessPoint[] ap = null;

  /** Das Rechteck das beim verschieben der Gruppe angezeigt wird. */
  private Rectangle2D drgRectangle = null;

  /** Erzeugt eine neue lose zusammenhngende Gruppe. */
  public ImplicitGroup() {
    setLinePaint(Color.yellow);
  }

  /** Fgt der Gruppe ein Selectable Objekt hinzu.
    * @param s Das Selectable, das der gruppe hinzugefgt werden soll.
    */
  public void add(Selectable s) {
    group.add(s);
    if (s instanceof Transformable) {
      support = support & ((Transformable) s).isGroupTransformationSupported();
    } else support = false;
    // if (!support) System.err.println("AT LEAST ONE GROUP MEMBER IS NOT GROUPABLE!!!");
    setShape(computeBounds());
  }

  /** Berechnet die BoundingBox der Gruppe.
    * @return Das die Gruppe umschlieende Rechteck.
    */
  private Rectangle2D computeBounds() {
    double min_x, min_y, max_x, max_y;
    // Liste aller Kindknoten einlesen und in ein Selectable
    if (group.size() >= 1) {
      Selectable firstMember = (Selectable) group.get(0);
      // Die Grenzen des ersten Kindes als minimale Grenzen setzen.
      if (firstMember != null) bounds = firstMember.getBounds2D();
      min_x = bounds.getMinX();
      min_y = bounds.getMinY();
      max_x = bounds.getMaxX();
      max_y = bounds.getMaxY();
      // alle
      for (int i=1; i<group.size(); i++) {
        try {
          Selectable child = (Selectable) group.get(i);
          if (child != null) {
            Rectangle rect = child.getBounds();
            if (rect.getMinX() < min_x ) min_x = rect.getMinX();
            if (rect.getMinY() < min_y ) min_y = rect.getMinY();
            if (rect.getMaxX() > max_x ) max_x = rect.getMaxX();
            if (rect.getMaxY() > max_y ) max_y = rect.getMaxY();
          }
        } catch (ClassCastException exc) {
          System.err.println("Object in Group wasn't a Selectable.");
          exc.printStackTrace();
        }
      }
      bounds = new Rectangle2D.Double(min_x, min_y, max_x-min_x, max_y-min_y);
    }
    return bounds;
  }

  /** Ruft fr alle Gruppenmitglieder rekursiv die handleGroupTransformation(..)
    * Methode auf. Als Parameter werden die Affine Transformation und die
    * BoundingBox der Gruppe bergeben.
    * @param at Die anzuwendende Affine Transformation.
    */
  public void setAffineTransform(AffineTransform at) {
    System.err.println("AT: " + at);
    if (!support) {
      return;
    }
    // super.setAffineTransform(at);
    computeBounds();
    for (int i=0; i<group.size(); i++) {
      Object obj = group.get(i);
      if (obj instanceof Transformable) {
        ((Transformable) group.get(i)).handleGroupTransformation(at, getBounds2D());
        System.err.println("calling recurse transformation");
      }
    }
    computeBounds();
  }

  /** Liefert die BoundingBox der Gruppe.
    * @return Die BoundingBox.
  public Rectangle2D getBounds2D() {
    return bounds;
  }

  /** Liefert die AccessPoints die fr Gruppentransformationen zur Verfgung
    * gestellt werden.
    * @param usage Der Verwendungszweck der AccessPoints.
    * @return Die AccessPoints fr den angegebenen Zweck.
    */
  public AbstractAccessPoint[] getAccessPoints(int usage) {
      switch (usage) {
        case UserDrawAction.SELECTMODE:
          if (ap == null) {
            // neue APs anlegen....
            ap = new SimpleAccessPoint[8];
            ap[0] = new SimpleAccessPoint(AbstractAccessPoint.SIZE_NW);
            ap[1] = new SimpleAccessPoint(AbstractAccessPoint.SIZE_N);
            ap[2] = new SimpleAccessPoint(AbstractAccessPoint.SIZE_NE);
            ap[3] = new SimpleAccessPoint(AbstractAccessPoint.SIZE_E);
            ap[4] = new SimpleAccessPoint(AbstractAccessPoint.SIZE_SE);
            ap[5] = new SimpleAccessPoint(AbstractAccessPoint.SIZE_S);
            ap[6] = new SimpleAccessPoint(AbstractAccessPoint.SIZE_SW);
            ap[7] = new SimpleAccessPoint(AbstractAccessPoint.SIZE_W);
          }
          alignSelectPoints();
          break;
        default:
          ap=null;
      }

    return ap;
    //TODO: Implement this jdraw.Transformable method
  }

  /** Positioniert die AccessPoints. */
  private void alignSelectPoints() {
    if (ap != null && ap.length >= 8) {
      double x = getBounds2D().getX();
      double y = getBounds2D().getY();
      double w = getBounds2D().getWidth();
      double h = getBounds2D().getHeight();

      Rectangle2D.Double r0 = new Rectangle2D.Double(x, y, 4, 4);
      Rectangle2D.Double r1 = new Rectangle2D.Double(x + w/2, y, 4, 4);
      Rectangle2D.Double r2 = new Rectangle2D.Double(x + w, y, 4, 4);
      Rectangle2D.Double r3 = new Rectangle2D.Double(x + w, y + h/2, 4, 4);
      Rectangle2D.Double r4 = new Rectangle2D.Double(x + w, y + h, 4, 4);
      Rectangle2D.Double r5 = new Rectangle2D.Double(x + w/2, y + h, 4, 4);
      Rectangle2D.Double r6 = new Rectangle2D.Double(x, y + h, 4, 4);
      Rectangle2D.Double r7 = new Rectangle2D.Double(x, y + h/2, 4, 4);

      ap[0].setAppearance(r0);
      ap[1].setAppearance(r1);
      ap[2].setAppearance(r2);
      ap[3].setAppearance(r3);
      ap[4].setAppearance(r4);
      ap[5].setAppearance(r5);
      ap[6].setAppearance(r6);
      ap[7].setAppearance(r7);
    }
  }

  /** Liefert die AccessPoints fr den SELECTMODE.
    * @return Die AccessPoints fr den SELECTMODE. */
  public AbstractAccessPoint[] getAccessPoints() {
    return getAccessPoints(UserDrawAction.SELECTMODE);
    //TODO: Implement this jdraw.Transformable method
  }

  /** Gibt die Resourcen die von den AccessPoints verbraucht werden wieder frei.*/
  public void freeAccessPoints() {
    ap = null;
    //TODO: Implement this jdraw.Transformable method
  }

  /** Behandelt das Ereignis, das ein AccessPoint gedrckt wurde.
    * @param e Das MouseEvent des Ereignisses.
    * @param ap Der AccessPoint, der gedrckt wurde.
    */
  public void handleAccessPointPressed(MouseEvent e, AbstractAccessPoint ap) {
    startPoint = e.getPoint();
    //TODO: Implement this jdraw.Transformable method
  }

  /** Behandelt das Ereignis, das ein AccessPoint gezogen wurde.
    * @param e Das MouseEvent des Ereignisses.
    * @param ap Der AccessPoint, der gezogen wurde.
    */
  public void handleAccessPointDragged(MouseEvent e, AbstractAccessPoint ap) {
    //TODO: Implement this jdraw.Transformable method
  }

  /** Behandelt das Ereignis, das ein AccessPoint losgelassen wurde.
    * @param e Das MouseEvent des Ereignisses.
    * @param ap Der AccessPoint, der losgelassen wurde.
    */
  public void handleAccessPointReleased(MouseEvent e, AbstractAccessPoint ap) {
    //TODO: Implement this jdraw.Transformable method
  }

  /** Behandelt das Ereignis, das die lose Gruppe als Element einer Gruppe
    * transformiert wurde.
    * @param at Die affine Transformation.
    * @param groupBounds Die BoundingBox der Gruppe.
    */
  public void handleGroupTransformation(AffineTransform at, Rectangle2D groupBounds) {
    //TODO: Implement this jdraw.Transformable method
  }

  /** Behandelt das Ereignis, das die Gruppe verschoben wird.
    * @param e Das MouseEvent des Ereignisses.
    * @param startPoint Die Position an der die Verschiebung begonnen hat.
    */
  public void handleDrag(MouseEvent e, Point2D startPoint) {
    if (drgRectangle == null) drgRectangle = new Rectangle2D.Double();
    double x,y,w,h;
    x = getBounds2D().getX() + e.getX() - startPoint.getX();
    y = getBounds2D().getY() + e.getY() - startPoint.getY();
    w = getBounds2D().getWidth();
    h = getBounds2D().getHeight();
    drgRectangle.setRect(x,y,w,h);

    for (int i=0; i<group.size(); i++) {
      Object obj = group.get(i);
      if (obj instanceof Transformable) {
        ((Transformable) group.get(i)).handleDrag(e, startPoint);
      }
    }

    //TODO: Implement this jdraw.Transformable method
  }

  /** Behandelt das Ereignis, das die Gruppe nach dem Verschieben wieder
    * losgelassen wurde.
    * @param e Das MouseEvent, das dieses Ereignis erzeugte.
    * @param startPoint Die Position, an der die Verschiebung begonnen hat.
    */
  public void handleDrop(MouseEvent e, Point2D startPoint) {
    drgRectangle = null;

    for (int i=0; i<group.size(); i++) {
      Object obj = group.get(i);
      if (obj instanceof Transformable) {
        ((Transformable) group.get(i)).handleDrop(e, startPoint);
      }
    }

    computeBounds();
  }

  /** Gibt an, ob dieses Transformable sich in eine Gruppe eingliedern lt.
    * @return <tt>false</tt> Da sich eine Implicitgroup nicht innerhalb einer
    * Gruppe befinden kann.
    */
  public boolean isGroupTransformationSupported() {
    return false;
    //TODO: Implement this jdraw.Transformable method
  }

  /** Stellt das Shape zur Verfgung, das gezeichnet wird, wenn die Gruppe
    * Verschoben wird.
    * @return Das DragShape.
    */
  public Shape supportDragShape() {
    return drgRectangle;
    //TODO: Implement this jdraw.Transformable method
  }

  /** Stellt das shape zur Vergung, das gezeichnet wird, wenn die Gruppe
    * ausgewhlt ist.
    * @param Das SelectedShape.
    */
  public Shape supportSelectedShape() {
    return getBounds2D();
    //TODO: Implement this jdraw.Transformable method
  }

}