/*******************************************************************************
  *
  *  program     : UserDrawAction.java
  *  author      : Helge Janicke
  *  version     : 2000-10-19(001)
  *  (c)         : Helge Janicke, Karsten Wolke, Niels-Peter de Witt
  *  email       : heljanic@agent.fho-emden.de
  *  description :
  ******************************************************************************
  * modifications:
  * date: ******* who: ****** description: *************************************
  * ----------------------------------------------------------------------------
  * 2000-10-19    heljanic
  * ----------------------------------------------------------------------------
  ******************************************************************************
  * comments:
  *
  ******************************************************************************
  * license:
  *
*******************************************************************************/

package jdraw;

import javax.swing.event.MouseInputAdapter;
import javax.swing.tree.*;
import javax.swing.JTree;
import javax.swing.JLabel;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.MouseEvent;
import java.util.Enumeration;

/** Diese Klasse dient als MouseAdapter.
  * Dieser behandelt die Mouseereignisse, die die GlassPane liefert.
  * Diese Klasse kann in verschiedenen Modi arbeiten.
  * Im SELECTMODE werden die Mouseereignisse geprft, ob sie ein spezielles
  * Zeichenobjekt treffen. Sollte dies der Fall sein wird je nach Ereignis
  * Dieses Objekt in die Liste der ausgewhlten Elemente aufgenommen. Wenn
  * es schon in dieser Liste existiert und nicht eine der Steuerungstasten gedrckt wurden, so wird es wieder aus der Liste entfernt. Wird ein bereits selektiertes Objekt gezogen, so wird geprft, ob dieses Objekt Transformationen untersttzt. Sollte dies der Fall sein wird es verschoben. hnliche Vorgehenweisen werden fr ROTATION und SHEAR Modus implementiert werden. Der DRAWMODE setzt voraus, das ein drawable Objekt definiert wurde. Befindet sich die Klasse UserDrawAction im DRAWMODE werden alle Mouseereignisse an die entsprechenden handle...() Methoden der Drawable Klasse deligiert. Sobald eine der handle...() Methoden den boolschen Wert <tt>true</tt> zurckliefert gilt die Zeichenoperation als abgeschlossen. Es wird von der Drawable Klasse eine Instanz des generierten Zeichenobjektes angefordert, die in der aktuellen Ebene als rightmost child eingefgt wird. Danach werden die Mousereignisse wieder an die Drawableklasse deligiert - bis zu einer Modus nderung.
  */
public class UserDrawAction extends MouseInputAdapter {

  public final static int SELECTMODE = 0;
  public final static int ROTATEMODE = 1;
  public final static int SHEARMODE  = 2;
  public final static int DRAWMODE   = 3;

  private static Rectangle2D mouseRectangle = new Rectangle2D.Double(0,0,8,8);

  private int mode = SELECTMODE;
  private Drawable drawable = null;
  private DefaultMutableTreeNode node = null;
  private AbstractAccessPoint currentAccessPoint = null;
  private Cursor currentCursor = Cursor.getDefaultCursor();
  private boolean canDrag = false;
  private boolean dragged = false;
  private Point2D startPoint = new Point2D.Double();
  private Selectable remOnRelease = null;
  private Selectable clearAndAddOnRelease = null;
  private DefaultTreeModel treeModel = null;
  private JTree tree;
  private JLabel lblPosition;

  /** Standartkonstruktor. */
  public UserDrawAction() {
  }

  /** Erzeugt eine neue UserDrawAction.
    * @param node Der Knoten an dem Eingefgt wird.
    * @param tree Der Baum, in dem eingefgt wird.
    * Der Baum wird bentigt um die graphischen Darstellung ber das TreeModel
    * ber nderungen zu informieren.
    */
  public UserDrawAction(DefaultMutableTreeNode node, JTree tree) {
    this();
    this.tree = tree;
    this.treeModel = (DefaultTreeModel) tree.getModel();
    setNode(node);
  }

  public void mousePressed(MouseEvent e) {
    GlassPane glasspane = null;
    if (e.getSource() instanceof GlassPane) glasspane = (GlassPane) e.getSource();
    // wenn nicht klar ist wohin gezeichnet werden kann, bzw worauf brechen wir ab.
    if (getNode() == null || glasspane == null) return;

    // Solange nichts angeklickt wurde, kann man auch nichts ziehen.
    // canDrag = false;
    startPoint.setLocation(e.getPoint());
    switch (getMode()) {
      case SELECTMODE:
        AbstractAccessPoint ap;
        Selectable selectable;
        ap = hitsAccessPoint(e);
        if (ap != null) {
          // Benutzer hat die mouse auf einem AbstractAccessPoint gedrckt.
          setCurrentAccessPoint(ap);
          Selectable sel = glasspane.getSelectable();
          if (sel instanceof Transformable) {
            // Eine Sicherheitsberprfung, da es eigentlich ein Transformable sein mu
            ((Transformable) sel).handleAccessPointPressed(e, ap);
          } else {
            System.err.println("UserDrawAction: \n\tEs waren AccessPoints vorhanden, obwohl die Auswahl nicht Transformable war!");
          }
        } else if ((selectable = hitsSelectable(e)) != null) {
          // Benutzer hat ein Selectable angeklickt.
          // und kann die Auswahl also auch bewegen.
          // canDrag = true;
          if (e.isShiftDown() || e.isControlDown() ) {
            // mehrfache Auswahl erlauben.
            if (glasspane.isSelected(selectable)) {
              // ist bereits ausgewhlt -> Aus Auswahl entfernen.
              remOnRelease = selectable;
            } else {
              // ist noch nicht ausgewhlt -> auswhlen.
              glasspane.addSelection(Color.blue);
//              System.err.println("adding to selection: "+ selectable);
              glasspane.addSelection(selectable);
            }
          } else {
            if (glasspane.isSelected(selectable)) {
              // nur einfache Auswahl.
              // und schon ausgewhlt.
              clearAndAddOnRelease = selectable;
            } else {
              // nur einfache Auswahl.
              // noch nicht ausgewhlt.
              glasspane.clearSelection();
              glasspane.addSelection(Color.blue);
  //            System.err.println("adding to selection: "+ selectable);
              glasspane.addSelection(selectable);
            }
          }
        } else {
          // Benutzer hat nichts angeklickt -> Auswahl aufheben wenn nicht Shift || Ctrl
          if (!e.isShiftDown() && !e.isControlDown()) {
            // Auswahl aufheben.
            glasspane.clearSelection();
          }
        }

        break;
      case ROTATEMODE:
        System.err.println("ROTATEMODE NI!");
        break;
      case SHEARMODE:
        System.err.println("SHEARMODE NI!");
        break;
      case DRAWMODE:
        if (getDrawable() != null) {
          if (getDrawable().handleMousePressed(e)) {
            // Es sind gengend Informationen um das Objekt zu erzeugen.
            add(glasspane);
          }
        } else {
          System.err.println("UserDrawAction:\n\tgetMode() == DRAWMODE && getDrawable() == null");
        }
        break;
      default:
        System.err.println("UserDrawAction: \n\tNOT A DEFINED DRAWING MODE!!!");
    }
    glasspane.refresh();
  }

  public void mouseDragged(MouseEvent e) {
    GlassPane glasspane = null;
    if (e.getSource() instanceof GlassPane) glasspane = (GlassPane) e.getSource();
    // wenn nicht klar ist wohin gezeichnet werden kann, bzw worauf brechen wir ab.
    if (getNode() == null || glasspane == null) return;

    switch (getMode()) {
      case SELECTMODE:
        AbstractAccessPoint ap;
        Selectable selectable;
        ap = getCurrentAccessPoint();
        remOnRelease = null;
        clearAndAddOnRelease = null;
        if (ap != null) {
          // Benutzer hat einen AccessPoint gedragged.

          Selectable sel = glasspane.getSelectable();
          if (sel instanceof Transformable) {
            // Eine Sicherheitsberprfung, da es eigentlich ein Transformable sein mu
            ((Transformable) sel).handleAccessPointDragged(e, ap);
          } else {
            System.err.println("UserDrawAction: \n\tEs waren AccessPoints vorhanden, obwohl die Auswahl nicht Transformable war!");
          }
        } else if ((selectable = glasspane.getSelectable()) != null) {
          // Benutzer hat ein Selectable gedragged.

          if (selectable instanceof Transformable) {
            // Eine Sicherheitsberprfung, da es eigentlich ein Transformable sein mu
            dragged = true;
            ((Transformable) selectable).handleDrag(e, startPoint);
          } else {
            System.err.println("UserDrawAction: \n\tEs wurde versucht ein Selectable zu draggen, obwohl es nicht Transformable war!");
          }
        }
        break;
      case ROTATEMODE:
        System.err.println("ROTATEMODE NI!");
        break;
      case SHEARMODE:
        System.err.println("SHEARMODE NI!");
        break;
      case DRAWMODE:
        if (getDrawable() != null) {
          if (getDrawable().handleMouseDragged(e)) {
            // Es sind gengend Informationen um das Objekt zu erzeugen.
            add(glasspane);
          }
        } else {
          System.err.println("UserDrawAction:\n\tgetMode() == DRAWMODE && getDrawable() == null");
        }
        break;
      default:
        System.err.println("UserDrawAction: \n\tNOT A DEFINED DRAWING MODE!!!");
    }
    glasspane.refresh();
    setPositionInfo(e);
  }

  public void mouseEntered(MouseEvent e) {
    GlassPane glasspane = null;
    if (e.getSource() instanceof GlassPane) glasspane = (GlassPane) e.getSource();

    // wenn nicht klar ist wohin gezeichnet werden kann, bzw worauf brechen wir ab.
    if (getNode() == null || glasspane == null) return;

    glasspane.setCursor(currentCursor);
  }

  public void mouseExited(MouseEvent e) {
  }

  public void mouseReleased(MouseEvent e) {
    GlassPane glasspane = null;
    if (e.getSource() instanceof GlassPane) glasspane = (GlassPane) e.getSource();
    // wenn nicht klar ist wohin gezeichnet werden kann, bzw worauf brechen wir ab.
    if (getNode() == null || glasspane == null) return;

    switch (getMode()) {
      case SELECTMODE:
        AbstractAccessPoint ap;
        Selectable selectable;
        ap = getCurrentAccessPoint();
        if (ap != null) {
          // Benutzer hat einen AccessPoint gedragged.

          Selectable sel = glasspane.getSelectable();
          if (sel instanceof Transformable) {
            // Eine Sicherheitsberprfung, da es eigentlich ein Transformable sein mu
            ((Transformable) sel).handleAccessPointReleased(e, ap);
          } else {
            System.err.println("UserDrawAction: \n\tEs waren AccessPoints vorhanden, obwohl die Auswahl nicht Transformable war!");
          }
        } else {
          if (remOnRelease != null) {
            glasspane.remSelection(remOnRelease);
          }
          if (clearAndAddOnRelease != null) {
            glasspane.clearSelection();
            glasspane.addSelection(Color.blue);
            glasspane.addSelection(clearAndAddOnRelease);
          }
          if (dragged && (selectable = glasspane.getSelectable()) != null) {
            if (selectable instanceof Transformable) {
              // Eine Sicherheitsberprfung, da es eigentlich ein Transformable sein mu
              ((Transformable) selectable).handleDrop(e, startPoint);
            }
          }
          clearAndAddOnRelease = null;
          remOnRelease = null;
          dragged = false;
        }
        break;
      case ROTATEMODE:
        System.err.println("ROTATEMODE NI!");
        break;
      case SHEARMODE:
        System.err.println("SHEARMODE NI!");
        break;
      case DRAWMODE:
        if (getDrawable() != null) {
          if (getDrawable().handleMouseReleased(e)) {
            // Es sind gengend Informationen um das Objekt zu erzeugen.
            add(glasspane);
          }
        } else {
          System.err.println("UserDrawAction:\n\tgetMode() == DRAWMODE && getDrawable() == null");
        }
        break;
      default:
        System.err.println("UserDrawAction: \n\tNOT A DEFINED DRAWING MODE!!!");
    }

    // Da nun egal was losgelassen wurde, knnen wir den currentAccessPoint wieder null setzen
    setCurrentAccessPoint(null);
    canDrag = false;
    glasspane.refresh();
    tree.updateUI();
  }

  public void mouseClicked(MouseEvent e) {
    GlassPane glasspane = null;
    if (e.getSource() instanceof GlassPane) glasspane = (GlassPane) e.getSource();

    // wenn nicht klar ist wohin gezeichnet werden kann, bzw worauf brechen wir ab.
    if (getNode() == null || glasspane == null) return;

    // Solage nichts angeklickt wurde, kann man auch nichts ziehen.
    canDrag = false;

    switch (getMode()) {
      case SELECTMODE:
        AbstractAccessPoint ap;
        Selectable selectable;
        ap = hitsAccessPoint(e);
        if (ap != null) {
          // Benutzer hat einen AccessPoint angeklickt.

          Selectable sel = glasspane.getSelectable();
          if (sel instanceof Transformable) {
            // Eine Sicherheitsberprfung, da es eigentlich ein Transformable sein mu
          } else {
            System.err.println("UserDrawAction: \n\tEs waren AccessPoints vorhanden, obwohl die Auswahl nicht Transformable war!");
          }
        } else if ((selectable = hitsSelectable(e)) != null) {
          // Benutzer hat ein Selectable angeklickt.

        } else {
          // Benutzer hat nichts angeklickt.

        }

        break;
      case ROTATEMODE:
        System.err.println("ROTATEMODE NI!");
        break;
      case SHEARMODE:
        System.err.println("SHEARMODE NI!");
        break;
      case DRAWMODE:
        if (getDrawable() != null) {
          if (getDrawable().handleMouseClicked(e)) {
            // Es sind gengend Informationen um das Objekt zu erzeugen.
            add(glasspane);
          }
        } else {
          System.err.println("UserDrawAction:\n\tgetMode() == DRAWMODE && getDrawable() == null");
        }
        break;
      default:
        System.err.println("UserDrawAction: \n\tNOT A DEFINED DRAWING MODE!!!");
    }
    glasspane.refresh();

  }


  public void mouseMoved(MouseEvent e) {

    GlassPane glasspane = null;
    if (e.getSource() instanceof GlassPane) glasspane = (GlassPane) e.getSource();

    // wenn nicht klar ist wohin gezeichnet werden kann, bzw worauf brechen wir ab.
    if (getNode() == null || glasspane == null) return;

    switch (getMode()) {
      case SELECTMODE:
        break;
      case ROTATEMODE:
        System.err.println("ROTATEMODE NI!");
        break;
      case SHEARMODE:
        System.err.println("SHEARMODE NI!");
        break;
      case DRAWMODE:
        if (getDrawable() != null) {
          if (getDrawable().handleMouseMoved(e)) {
            // Es sind gengend Informationen um das Objekt zu erzeugen.
            add(glasspane);
          }
          glasspane.refresh();
        } else {
          System.err.println("UserDrawAction:\n\tgetMode() == DRAWMODE && getDrawable() == null");
        }
        break;
      default:
        System.err.println("UserDrawAction: \n\tNOT A DEFINED DRAWING MODE!!!");
    }
    setPositionInfo(e);
  }

  /** Liefert den aktuellen Modus zurck.
    * @return den aktuellen Modus.
    * @see SELECTMODE
    * @see DRAWMODE
    * @see ROTATEMODE
    * @see SHEARMODE
    */
  public int getMode() {
    return mode;
  }

  /** Setzt den aktuellen Modus.
    * @param mode der neue Modus.
    * @see SELECTMODE
    * @see DRAWMODE
    * @see ROTATEMODE
    * @see SHEARMODE
    */
  public void setMode(int mode) {
    if (this.mode == DRAWMODE && getDrawable() != null) {
      // Wir waren im Zeichenmodus. Dem vielleicht im Zeichnen befindlichen
      // Objekt die mglichkeit geben sich fertigzustellen.
      getDrawable().handleTermination();
      // add(glasspane);
    }
    this.mode = mode;
    currentCursor = Cursor.getDefaultCursor();
  }

  /** Liefert den aktuellen Knoten, an dem neue Objekte eingehngt werden.
    * @return den aktuellen Knoten.
    */
  public DefaultMutableTreeNode getNode() {
    return node;
  }

  /** Setzt den aktuellen Knoten, an dem neue Objekte eingefgt werden.
    * @param der neue Knoten.
    */
  public void setNode(DefaultMutableTreeNode node) {
    this.node = node;
  }

  /** Liefert das aktuelle Drawable, das benutzt wird um neue Objekte zu zeichnen.
    * @return Das aktuelle Drawable.
    */
  public Drawable getDrawable() {
    return drawable;
  }

  /** Setzt das aktuelle Drawable, das benutzt wird um neue Objekte zu zeichnen.
    * @param drawable Das neue Drawable.
    */
  public void setDrawable(Drawable drawable) {
    this.drawable = drawable;
    currentCursor = drawable.supportCursor();
  }

  /** Setzt den Modus auf DRAWMODE und das neue Drawable.
    * @param drawable Das neue Drawable.
    */
  public void setDrawMode(Drawable drawable) {
    setMode(DRAWMODE);
    setDrawable(drawable);
    System.err.println("Setting drawmode");
  }

  /** Fgt das AbstractGraphObject als neues Blatt an den Knoten node an.
    * @param o Das AbstractGraphObject, das hizugefgt werden soll.
    */
  private void add(AbstractGraphObject o) {
    // DefaultMutableTreeNode aNode = new DefaultMutableTreeNode(o);
    addObject(getNode(), o);
  }

  /** Versucht von dem aktuellen Drawable eine Instanz zu erhalten und fgt diese
    * dann im Baum ein.
    * @see add(AbstractGraphObject o)
    */
  private void add(GlassPane gp) {
    if (getDrawable() != null) {
      AbstractGraphObject obj = getDrawable().getInstance();
      if (obj != null) {
        add(obj);
        gp.clearSelection();
        gp.addSelection(Color.blue);
        gp.addSelection(obj);
      }
    }
  }

  /** Prft ob einer der AccessPoints der aktuellen Auswahl angeklickt wurde.
    * @param p Die Mouseposition, die berprft werden soll.
    * @return Den angeklickten AccessPoint oder <tt>null</tt>.
    */
  private AbstractAccessPoint hitsAccessPoint(MouseEvent e) {
    GlassPane glasspane = null;
    boolean intersects = false;

    if (e.getSource() instanceof GlassPane) glasspane = (GlassPane) e.getSource();
    // wenn nicht klar ist wohin gezeichnet werden kann, bzw worauf brechen wir ab.
    if (getNode() == null || glasspane == null) return null;

    adjustMouseRectangle(e);

    if (glasspane.getSelectable() instanceof Transformable) {
      AbstractAccessPoint[] ap = ((Transformable) glasspane.getSelectable()).getAccessPoints(getMode());
      if (ap != null) {
        for (int i=0; i<ap.length; i++) {
          if (ap[i].getAppearance().intersects(mouseRectangle)) return ap[i];
        }
      }
    }

    return null;
  }

  /** Prft ob ein Selectable angeklickt wurde. Der Scope dieser Funktion ist:
    * alle Selectables die Kinder von node sind.
    * @param p Die Mouseposition, die berprft werden soll.
    * @return Das angeklickte Selectable oder <tt>null</tt>.
    */
  private Selectable hitsSelectable(MouseEvent e) {
    GlassPane glasspane = null;
    boolean intersects = false;

    if (e.getSource() instanceof GlassPane) glasspane = (GlassPane) e.getSource();
    // wenn nicht klar ist wohin gezeichnet werden kann, bzw worauf brechen wir ab.
    if (getNode() == null || glasspane == null) return null;

    adjustMouseRectangle(e);

    for (Enumeration enum = getNode().preorderEnumeration(); enum.hasMoreElements(); ) {

      Object nextnode = enum.nextElement();

      Object obj = ((DefaultMutableTreeNode) nextnode).getUserObject();

      if (obj instanceof AbstractGraphObject) {
        // ein GraphObject und alles wird gut...
        AbstractGraphObject grph = (AbstractGraphObject) obj;
        if (grph.isFilled()) {
          intersects = grph.intersects(mouseRectangle);
        } else {
          intersects = AbstractGraphObject.intersectsBoundary(grph, mouseRectangle);
        }
      } else if (obj instanceof Selectable) {
        // Wer macht denn so was?! aber na ja geht auch...
        intersects = ((Selectable) obj).intersects(mouseRectangle);
      }

      if (intersects) {
        return (Selectable) obj;
      }
    }

    return null;
  }

  private void adjustMouseRectangle(MouseEvent e) {
    mouseRectangle.setRect(e.getX() - (mouseRectangle.getWidth()/2),
        e.getY() - (mouseRectangle.getHeight()/2), mouseRectangle.getWidth(),
        mouseRectangle.getHeight());
  }


  /** Liefert den currentAccessPoint.
    * @return den currentAccessPoint.
    */
  private AbstractAccessPoint getCurrentAccessPoint() {
    return currentAccessPoint;
  }

  /** Setzt den currentAccessPoint.
    * @param ap der neue AccessPoint.
    */
  private void setCurrentAccessPoint(AbstractAccessPoint ap) {
    this.currentAccessPoint = ap;
  }


  /** Fgt ein neues Objekt in den Baum ein.
    * @param parentnode Der Knoten an dem das Object eingefgt werden soll.
    * @param child Das UserObject (AbstractGraphObject, Group ...) das in dem neuen Kindknoten stehen soll.
    */
  public DefaultMutableTreeNode addObject(DefaultMutableTreeNode parentNode, Object child) {
    DefaultMutableTreeNode childNode = new DefaultMutableTreeNode(child);
    childNode.setAllowsChildren(false);
    treeModel.insertNodeInto(childNode, parentNode, parentNode.getChildCount());
    // Make sure the user can see the lovely new node.
    tree.scrollPathToVisible(new TreePath(childNode.getPath()));
    return childNode;
  }

  /** Setzt das Label, in das die Informationen ber die Mouseposition geschrieben werden sollen.
    * @param l Das JLabel.
    */
  public void setPositionLabel(JLabel l) {
    this.lblPosition = l;
  }

  /** Schreibt die Mouseposition in das Positionslabel.
    * @param e Das Mouseevent.
    */
  private void setPositionInfo(MouseEvent e) {
    if (lblPosition != null) {
      this.lblPosition.setText("( int x ; int y)");
    }
  }
}